local colors = require 'colors'

local urutora = {
  nodes = {},
  stateStack = {},
  font = love.graphics.newFont(),
  defaults = {
    cb = function() end,
    w = 100,
    sliderValue = 0.5,
  },
  nodeTypes = {
    LABEL = 1,
    BUTTON = 2,
    SLIDER = 3,
  },
  mouseButtons = {
    LEFT = 1,
  },
}

urutora.style = {
  padding = urutora.font:getHeight() / 2,
  bgColor = colors.LOVE_BLUE,
  hoverBgColor = colors.WHITE,
  fgColor = colors.WHITE,
  hoverFgColor = colors.LOVE_BLUE,
  howRounded = 0.25
}

---------------------------------------------------------------------------
---------------------------------------------------------------------------

local function getMouse()
  return love.mouse.getX(), love.mouse.getY()
end

local function saveState()
  local s = {}
  s.color = { love.graphics.getColor() }
  s.lineWidth = love.graphics.getLineWidth()
  s.lineStyle = love.graphics.getLineStyle()
  s.font = love.graphics.getFont()

  table.insert(urutora.stateStack, s)
end

local function restoreState()
  local s = table.remove(urutora.stateStack, #urutora.stateStack)

  love.graphics.setColor(s.color)
  love.graphics.setLineWidth(s.lineWidth)
  love.graphics.setLineStyle(s.lineStyle)
  love.graphics.setFont(s.font)
end

local function restoretate()
  local s = {}
  s.color = love.graphics.getColor()
  s.lineWidth = love.graphics.getLineWidth()
  s.lineStyle = love.graphics.getLineStyle()
  s.font = love.graphics.getFont()

  table.remove(urutora.stateStack, #urutora.stateStack + 1, s)
end

local function getLayerColors(node)
  if node.pointed then
    return urutora.style.hoverBgColor, urutora.style.hoverFgColor
  end

  return urutora.style.bgColor, urutora.style.fgColor
end


local function drawBaseRectangle(node)
  local bgc, _ = getLayerColors(node)
  love.graphics.setColor(bgc)

  love.graphics.rectangle('fill',
    math.floor(node.x),
    math.floor(node.y),
    math.floor(node.w),
    math.floor(node.h),
    urutora.style.howRounded * node.h / 2,
    urutora.style.howRounded * node.h / 2
  )
end

local function drawText(node)
  if not node.text then
    return
  end

  local _, fgc = getLayerColors(node)
  love.graphics.setColor(fgc)
  
  love.graphics.print(
    node.text,
    math.floor(node.x + node.padding),
    math.floor(node.y + node.padding)
  )
end

local function drawBase(node)
  drawBaseRectangle(node)
  drawText(node)
end

local function isPointInsideNode(x, y, node)
  return not (
    x < node.x or
    x > (node.x + node.w) or
    y < node.y or
    y > (node.y + node.h)
  )
end

local function setBounds(node, x, y, w, h)
  local f = urutora.font
  node.padding = urutora.style.padding

  node.x = x
  node.y = y
  node.w = w or f:getWidth(node.text) + node.padding * 2
  node.h = h or f:getHeight() + node.padding * 2
end

local function initNode(node, data)
  node.text = data.text
  setBounds(
    node,
    data.x or 1,
    data.y or 1,
    data.w or urutora.defaults.w
  )
end

local function isLabel(node)
  return node.type == urutora.nodeTypes.LABEL
end

---------------------------------------------------------------------------
---------------------------------------------------------------------------

function urutora.newLabel(data)
  local node = {}
  data = data or {}
  initNode(node, data)

  function node:update() end
  function node:draw() end
  node.type = urutora.nodeTypes.LABEL
  node.callback = data.action or urutora.defaults.cb
  
  table.insert(urutora.nodes, node)
  return node
end

---------------------------------------------------------------------------
---------------------------------------------------------------------------

function urutora.newButton(data)
  local node = {}
  data = data or {}
  initNode(node, data)

  function node:update() end
  function node:draw() end
  node.type = urutora.nodeTypes.BUTTON
  node.callback = data.action or urutora.defaults.cb
  
  table.insert(urutora.nodes, node)
  return node
end

---------------------------------------------------------------------------
---------------------------------------------------------------------------

function urutora.newSlider(data)
  local node = {}
  data = data or {}
  initNode(node, data)

  function node:update() end
  function node:draw() end
  node.type = urutora.nodeTypes.SLIDER
  node.callback = data.action or urutora.defaults.cb

  node.value = urutora.defaults.sliderValue
  
  function node:draw()
    local _, fgc = getLayerColors(self)
    love.graphics.setColor(fgc)
    
    love.graphics.circle('line',
      self.x + self.padding + (self.w - self.padding * 2) * self.value,
      self.y + self.h / 2,
      self.padding
    )
  end
  
  function node:update(dt)
    local x, y = getMouse()

    if love.mouse.isDown(urutora.mouseButtons.LEFT) then
      self.value = (x - (self.x + self.padding)) * (self.w - self.padding * 2)
      if self.value > 1 then self.value = 1 end
      if self.value < 0 then self.value = 0 end
    end
  end
  
  table.insert(urutora.nodes, node)
  return node
end

---------------------------------------------------------------------------
---------------------------------------------------------------------------

function urutora.draw()
  saveState()
  
  for _, node in ipairs(urutora.nodes) do
    if not isLabel(node) then
      drawBaseRectangle(node)
    end

    drawText(node)
    node:draw()
  end

  restoreState()
end

function urutora.update(dt)
  for _, node in ipairs(urutora.nodes) do
    local x, y = getMouse()

    node.pointed = isPointInsideNode(x, y, node) and not isLabel(node)
    node:update(dt)
  end
end

function urutora.released(x, y)
  for _, node in ipairs(urutora.nodes) do
    if isPointInsideNode(x, y, node) then
      node.callback()
    end
  end
end

return urutora
